// MultiAperture_.java

import ij.*;
import ij.gui.*;
import ij.measure.*;
import ij.plugin.*;
import ij.process.*;
import ij.text.*;
import ij.gui.Toolbar.*; 

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.Canvas.*;
import java.util.*;

import astroj.*;
import javax.swing.JLabel;


/**
 * Based on Aperture_.java, but with pre-selection of multiple apertures and processing of stacks.
 * 
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @version 1.0
 * @date 2006-Feb-15
 *
 * @version 1.1
 * @date 2006-Nov-29
 * @changes Slight re-organization so that it's easier to use MultiAperture_ as a parent class, e.g. for Stack_Aligner_.
 *
 * @version 1.2
 * @date 2006-Dec-6
 * @changes Corrected problem with non-instantiated xOld for the case of images and not stacks;
 *	added thread to display current ROIs.
 *
 * @version 1.3
 * @date 2006-Dec-11
 * @changes Complicated thread solution replaced with simple overlay solution inherited from Aperture_
 *
 * @version 1.4
 * @date 2007-Mar-14
 * @changes Added <ESC> to stop processing of large stacks which have gone awry.
 *
 * @version 1.5
 * @date 2007-May-01
 * @changes Added output of ratio
 *
 * @version 1.6
 * @date 2010-Mar-18
 * @author Karen Collins (Univ. Louisville/KY)
 * @changes
 * 1) On my machine, the number of apertures and the 4 check box option values do not get stored between runs.
 *	I don't know if this is a problem with my computing environment, or if it is a problem in general.
 *	To make it work on my end, I had to implement the MultiAperture_ preferences retrieval from the Aperture_
 *	preferences management code. I was able to implement the save (set) from MultiAperture_ without problems.
 *	I could never work out how to do both from MultiAperture_, so my solution is probably not ideal, but it seems to work now.
 * 2) Implemented the Ratio Error functionality. The code I am currently using is:
 *		ratio*Math.sqrt(srcVar/(src*src)+
 *		othersVariance/(others*others))        {equation 2}
 *	where
 *		ratio = src counts/total of all comparison counts  (all adjusted for sky background)
 *		srcVar = (source error)^2 from the 1st of N apertures per image (as calculated in Photometer above)
 *		src = total source counts less sky background from the 1st of N apertures
 *		totVar = (ap#2 error)^2 + (ap#3 error)^2 + ...  + (ap#N error)^2
 *		tot = total source counts less sky background from the 2nd through Nth apertures
 * 3) Implemented the Ratio SNR functionality. It simply reports (src counts)/{equation 2}, which is accomplished by
 *	not multiplying by "ratio".
 * 4) Measurements table code to support the ratio error and ratioSNR reporting.
 *
 * @version 1.7
 * @date 2011-Mar-29
 * @author F. Hessman (Goettingen)
 * @changes added dialog() to make it easier to sub-class (e.g. Stack_Aligner)
 */
public class MultiAperture_ extends Aperture_ implements MouseListener, KeyListener
	{
	boolean cancelled=false;
	boolean verbose=true;
	boolean blocked=false;
	boolean previous=false;
	boolean doStack=false;
    boolean processingStack=false;
    boolean apertureRemoved=false;
    boolean apertureChanged=false;
    boolean firstClick = true;
    boolean enterPressed = false;
    
    int xLocation = 10, yLocation = 10;
	int firstSlice=1;
	int lastSlice=1;

    int astronomyToolId = 0;
    int currentToolId = 0;
    Toolbar toolbar;


	String infoMessage = new String("");

//	double vx = 0.0;
//	double vy = 0.0;
//	double vxOld = 0.0;
//	double vyOld = 0.0;
//	public static String VX = new String ("vx");
//	public static String VY = new String ("vy");

	protected int ngot=0;
	protected int aperture=0;
    protected int nAperturesMax=500;
    protected int nApertures=2;
    protected int nAperturesStored=0;
    protected int startDragScreenX;
    protected int startDragScreenY;
    
	protected double[] xOld;
	protected double[] yOld;
	protected double[] xPosStored;
	protected double[] yPosStored;
    protected boolean[] isRefStarStored;
	protected String xOldApertures, yOldApertures, isOldRefStar;    

	protected double[] xPos;
	protected double[] yPos;    
    protected double[] ratio;
    protected double[] ratioError;
    protected double[] ratioSNR;
    protected boolean[] isRefStar;    // indication of classification as reference star or target star (target stars not included in total ref star count)
	protected double[] src;           // net integrated counts for each aperture
	protected double[] srcVar;        // error*error for each aperture
	protected double[] tot;		    // total ref star counts for each source
    protected double[] totVar;        // variance in tot ref star counts
	protected double peak = 0.0;		// max pixel value in aperture
	
    double apFWHMFactor = 1.4;
    double xFWHM = 0.0;
    double yFWHM = 0.0;

    double oldradius;
    double oldrBack1;
    double oldrBack2;
    double oldapFWHMFactor;
    boolean oldUseVarSizeAp;
    boolean oldBackIsPlane;
    boolean oldRemoveBackStars;   

//	double ratio = 0.0;		// FIRST APERTURE
//	double ratioError = 0.0;
//	double ratioSNR = 0.0;
	public static String RATIO = new String ("rel_flux_T1");
    public static String TOTAL = new String ("tot_C_cnts");
    public static String TOTAL_ERROR = new String ("tot_C_err");

    protected boolean autoMode=false;
    protected boolean singleStep=false;
    protected boolean simulatedLeftClick=false;
    protected boolean aperturesInitialized=false;
    protected boolean allStoredAperturesPlaced=false;
    protected boolean enableDoubleClicks = true;
    protected boolean multiApertureRunning = false;
    protected boolean useVarSizeAp= false;
    protected boolean showHelp= true;
//	protected boolean follow=false;
//	protected boolean wideTable=true;

	protected boolean showRatio=true;
    protected boolean showCompTot=true;
	protected boolean showRatioError=true;
	protected boolean showRatioSNR=true;
    protected boolean updatePlot=true;
    protected boolean useMacroImage=false;
    protected String macroImageName=null;
    ImagePlus openImage;
    AstroStackWindow aw;
    

    protected boolean frameAdvance=false;

    MouseEvent dummyClick = null;
    MouseEvent ee;

    protected int screenX;
    protected int screenY;
    protected int modis;
    
    JFrame helpFrame;
    JScrollPane helpScrollPane; 
    JPanel helpPanel;
    JLabel leftClickLabel, shiftLeftClickLabel, shiftControlLeftClickLabel, controlLeftClickLabel, rightClickLabel, controlRightClickLabel;
    JLabel escapeLabel, enterLabel, mouseWheelLabel, middleClickLabel, leftClickDragLabel;  
    ImageIcon MAIcon;
    int helpFrameLocationX = 10;
    int helpFrameLocationY = 10;
    
    TimerTask stackTask = null;
    java.util.Timer stackTaskTimer = null;

    boolean doubleClick = false;
    TimerTask doubleClickTask = null;
    java.util.Timer doubleClickTaskTimer = null;
	
    protected static String PREFS_AUTOMODE        = new String ("multiaperture.automode");  //0 click - for use with macros
    protected static String PREFS_FINISHED        = new String ("multiaperture.finished");  //signals finished to macros
    protected static String PREFS_CANCELED        = new String ("multiaperture.canceled");
    protected static String PREFS_PREVIOUS        = new String ("multiaperture.previous");
    protected static String PREFS_SINGLESTEP      = new String ("multiaperture.singlestep");
    protected static String PREFS_USEVARSIZEAP    = new String ("multiaperture.usevarsizeap");
    protected static String PREFS_SHOWHELP        = new String ("multiaperture.showhelp");
    protected static String PREFS_APFWHMFACTOR    = new String ("multiaperture.apfwhmfactor");
//    protected static String PREFS_FOLLOW          = new String ("multiaperture.follow");
//	protected static String PREFS_WIDETABLE       = new String ("multiaperture.widetable");
	protected static String PREFS_SHOWRATIO       = new String ("multiaperture.showratio");
    protected static String PREFS_SHOWCOMPTOT     = new String ("multiaperture.showcomptot");
	protected static String PREFS_SHOWRATIO_ERROR = new String ("multiaperture.showratioerror");
	protected static String PREFS_SHOWRATIO_SNR   = new String ("multiaperture.showratiosnr");
    protected static String PREFS_NAPERTURESMAX   = new String ("multiaperture.naperturesmax");
	protected static String PREFS_XAPERTURES      = new String ("multiaperture.xapertures");
	protected static String PREFS_YAPERTURES      = new String ("multiaperture.yapertures");
    protected static String PREFS_ISREFSTAR       = new String ("multiaperture.isrefstar");
    protected static String PREFS_USEMACROIMAGE   = new String ("multiaperture.useMacroImage");
    protected static String PREFS_MACROIMAGENAME  = new String ("multiaperture.macroImageName");
    protected static String PREFS_ENABLEDOUBLECLICKS  = new String ("multiaperture.enableDoubleClicks");
    protected static String PREFS_UPDATEPLOT  = new String ("multiaperture.updatePlot");
    protected static String PREFS_XLOCATION  = new String ("multiaperture.xLocation");
    protected static String PREFS_YLOCATION  = new String ("multiaperture.yLocation");
    protected static String PREFS_PREVNUMMONITORS  = new String ("multiaperture.prevNumberOfMonitors");

//	public static double RETRY_RADIUS = 3.0;



	/**
	 * Standard ImageJ PluginFilter setup routine.
	 */
	public int setup (String arg, ImagePlus imp)
		{
        this.getMeasurementPrefs();
        if (useMacroImage)
                {
                openImage = WindowManager.getImage(macroImageName);
                imp = openImage;
                }
        if (imp == null)
            {
            cancel();
            return DONE;
            }

        if (imp.getWindow() instanceof PlotWindow)     //if plotwindow is selected and only one other image window is open
            {                                          //switch to other image if not a plotwindow, otherwise message to user
            int numImages=WindowManager.getImageCount();
            if (numImages == 1)
                {
                IJ.showMessage("No image windows open for Multi-aperture processing");
                cancel();
                return DONE;
                }
            if (numImages == 2)
                {
                int[] idList = WindowManager.getIDList();
                imp = idList[0] == imp.getID() ? WindowManager.getImage(idList[1]) : WindowManager.getImage(idList[0]);
                if (imp.getWindow() instanceof PlotWindow)
                    {
                    IJ.showMessage("No image windows open for Multi-aperture processing");
                    cancel();
                    return DONE;
                    }
                }
            else
                {
                IJ.showMessage("Select image window to process, then restart Multi-aperture");
                cancel();
                return DONE;
                }
            }

        // TO MAKE SURE THAT THE NEXT CLICK WILL WORK
        toolbar = Toolbar.getInstance();
        astronomyToolId = toolbar.getToolId("Astronomy_Tool");
        currentToolId = Toolbar.getToolId();
        if (currentToolId != astronomyToolId)
            {
            if (astronomyToolId != -1)
            	IJ.setTool(astronomyToolId);
            else
                IJ.setTool(0);
            }

		IJ.register(MultiAperture_.class);
		return super.setup(arg,imp);
		}

    
    /**
	 * Standard ImageJ Plug-in method which runs the plug-in, notes whether the image is a stack,
	 * and registers the routine for mouse clicks.
	 */
	public void run (ImageProcessor ip)
		{
//        if (table==null)
//            IJ.log("Table is null");
//        else
//            IJ.log("Table rows = "+table.getCounter());
        
        Frame[] Frames = JFrame.getFrames();
        if (Frames.length > 0)
            {
            for (int i = 0; i < Frames.length; i++)
                {
                if (Frames[i].getTitle().equals("Multi-Aperture Help"))
                    {
                    Frames[i].dispose();
                    }
                }
            }
      
        Prefs.set (MultiAperture_.PREFS_CANCELED, "false");
		this.getMeasurementPrefs();
        suffix="_T1";
        if (!autoMode)
            {
            String[] apsX = xOldApertures.split(",");
            String[] apsY = yOldApertures.split(",");
            if (apsX.length != apsY.length)
                nAperturesStored = 0;
            else
                nAperturesStored = apsX.length;
            }
        if (useMacroImage)
            {
            openImage = WindowManager.getImage(macroImageName);

            if (!(openImage == null))
                {
                imp = openImage;
                }
            }

		// GET HOW MANY APERTURES WILL BE MEASURED WITH WHAT RADII

		if ( !setUpApertures() || nApertures == 0 || !prepare () )
			{
            imp.unlock();
            cancel();
			return;
			}

		// START ESCAPE ABORTION POSSIBILITY
		IJ.resetEscape();

		// REGISTER FOR MOUSE CLICKS

        if ( !autoMode )
                {
                MouseListener[] ml = canvas.getMouseListeners();  
                for (int i = 0; i < ml.length; i++)
                    {
                    if (ml[i] instanceof MultiAperture_)
                        {
//                        IJ.log("MouseListener already running");
                        canvas.removeMouseListener(ml[i]);
                        }
                    }
                KeyListener[] kl = canvas.getKeyListeners();
                for (int i = 0; i < kl.length; i++)
                    {
                    if (kl[i] instanceof MultiAperture_)
                        {
//                        IJ.log("KeyListener already running");
                        canvas.removeKeyListener(kl[i]);
                        }
                    }
                canvas.addMouseListener(this);
                canvas.addKeyListener(this);
                }

		if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
			ocanvas.clearRois();
        if (previous)
            {
            infoMessage = "Please select first aperture (right click to finalize) ...";
            IJ.showStatus (infoMessage);
            }
        setApertureColor(Color.green);
        setApertureName("T1");

        if (imp.getWindow() instanceof astroj.AstroStackWindow)
            {
            aw = (AstroStackWindow)imp.getWindow();
            aw.setDisableShiftClick(true);
            }
//        imp.getWindow().requestFocus();
//        imp.getCanvas().requestFocusInWindow();

		if ( autoMode )
            mouseReleased( dummyClick );
        else
            {
            imp.getWindow().requestFocus();
            imp.getCanvas().requestFocusInWindow();
            }
        }

    protected void cancel()
        {
        if (table != null) table.setLock(false);
        Prefs.set (MultiAperture_.PREFS_AUTOMODE, "false");
        Prefs.set (MultiAperture_.PREFS_FINISHED, "true");
        Prefs.set (MultiAperture_.PREFS_USEMACROIMAGE, "false");
        Prefs.set (MultiAperture_.PREFS_CANCELED, "true");
        }

	/**
	 * Get all preferences.
	 */
	protected void getMeasurementPrefs()
		{
		super.getMeasurementPrefs();
        oldBackIsPlane = backIsPlane;
        oldRemoveBackStars = removeBackStars;
        
		autoMode       = Prefs.get (MultiAperture_.PREFS_AUTOMODE, autoMode);
        useMacroImage  = Prefs.get (MultiAperture_.PREFS_USEMACROIMAGE, useMacroImage);
        macroImageName = Prefs.get (MultiAperture_.PREFS_MACROIMAGENAME, macroImageName);
        previous       = Prefs.get (MultiAperture_.PREFS_PREVIOUS, previous);
        singleStep     = Prefs.get (MultiAperture_.PREFS_SINGLESTEP, singleStep);
//        wideTable      = Prefs.get (MultiAperture_.PREFS_WIDETABLE, wideTable);
//		follow         = false;	// Prefs.get (MultiAperture_.PREFS_FOLLOW,    follow);
        oldradius = radius;
        oldrBack1 = rBack1;
        oldrBack2 = rBack2;
		showRatio      = Prefs.get (MultiAperture_.PREFS_SHOWRATIO, showRatio);
        showCompTot    = Prefs.get (MultiAperture_.PREFS_SHOWCOMPTOT, showCompTot);
		showRatioError = Prefs.get (MultiAperture_.PREFS_SHOWRATIO_ERROR, showRatioError);
		showRatioSNR   = Prefs.get (MultiAperture_.PREFS_SHOWRATIO_SNR, showRatioSNR);
		useVarSizeAp   = Prefs.get (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
        oldUseVarSizeAp = useVarSizeAp;
		apFWHMFactor   = Prefs.get (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
        oldapFWHMFactor = apFWHMFactor;
        nAperturesMax  = (int) Prefs.get (MultiAperture_.PREFS_NAPERTURESMAX, nAperturesMax);
        enableDoubleClicks = Prefs.get (MultiAperture_.PREFS_ENABLEDOUBLECLICKS, enableDoubleClicks);
        showHelp  = Prefs.get (MultiAperture_.PREFS_SHOWHELP, showHelp);
		xOldApertures = Prefs.get (MultiAperture_.PREFS_XAPERTURES,"");
		yOldApertures = Prefs.get (MultiAperture_.PREFS_YAPERTURES,"");
        isOldRefStar = Prefs.get (MultiAperture_.PREFS_ISREFSTAR,"");
        updatePlot = Prefs.get (MultiAperture_.PREFS_UPDATEPLOT,updatePlot);
        xLocation = (int) Prefs.get (MultiAperture_.PREFS_XLOCATION, xLocation);
        yLocation = (int) Prefs.get (MultiAperture_.PREFS_YLOCATION, yLocation);
        helpFrameLocationX = (int) Prefs.get ("plot2.helpFrameLocationX", helpFrameLocationX);
        helpFrameLocationY = (int) Prefs.get ("plot2.helpFrameLocationY", helpFrameLocationY);
		}

	/**
	 * Initializes variables etc.
	 */
	protected boolean prepare ()
		{
//		if (!checkResultsTable ()) return false;  //removed to allow dynamic number of apertures using right click to terminate aperture selections
		// LOAD PREVIOUS APERTURES IF DESIRED
        isRefStar = new boolean[nAperturesMax];
		if ( previous )
			{
			String[] aps = xOldApertures.split(",");
			nAperturesStored = aps.length;
            xPosStored = extract(true, aps);
            
			aps = yOldApertures.split(",");
			if (nAperturesStored != aps.length)
				{
				IJ.error("The stored apertures are not consistent: "+nAperturesStored+"!="+aps.length);
				return false;
				}
            yPosStored = extract(false, aps);
            
			aps = isOldRefStar.split(",");
			if (nAperturesStored != aps.length || isInstanceOfStackAlign)
				{
                isRefStarStored = new boolean[nAperturesStored];
                for (int ap = 0; ap < nAperturesStored; ap++)
                    {
                    if (ap == 0)
                        isRefStarStored[ap] = false;
                    else
                        isRefStarStored[ap] = true;
                    }
				}
            else
                {
                isRefStarStored = extractBoolean(aps);  
                }
            int size = Math.min(nAperturesStored, nAperturesMax);
            for (int ap = 0; ap < size; ap++)
                {
                isRefStar[ap] = isRefStarStored[ap];
                }            
			}
		if ( autoMode )
			{
			String[] aps = xOldApertures.split(",");
			nApertures = aps.length;
            xPos = extract(true, aps);
            
			aps = yOldApertures.split(",");
			if (nApertures != aps.length)
				{
				IJ.error("The stored apertures are not consistent: "+nApertures+"!="+aps.length);
				return false;
				}
            aperturesInitialized = true;
            yPos = extract(false, aps);
            
			aps = isOldRefStar.split(",");
			if (nApertures != aps.length)
				{
                isRefStar = new boolean[nApertures];
                for (int ap = 0; ap < nApertures; ap++)
                    {
                    if (ap == 0)
                        isRefStar[ap] = false;
                    else
                        isRefStar[ap] = true;
                    }
				}
            else
                {
                isRefStar = extractBoolean(aps);
                }
			}
		else	{
            xPos = new double[nApertures];
            yPos = new double[nApertures];
			}
		if (xPos == null || yPos == null || isRefStar == null)
			{
			IJ.error("Null aperture arrays???");
			return false;
			}

		imp.setSlice(firstSlice);
		ip = imp.getProcessor();
		imp.killRoi();
		return true;
		}

	/**
	 * Extracts a double array from a string array.
	 */
	protected double[] extract (boolean XnotY, String[] s)
		{
        boolean isFITS = false;
		double[] arr = new double[s.length];
        if (s.length > 0 && s[0].startsWith("FITS"))
            {
            isFITS = true;
            s[0] = s[0].substring(4);
            }
		try	{
			for (int i=0; i < arr.length; i++)
				arr[i] = Double.parseDouble(s[i]);
			}
		catch (NumberFormatException e)
			{
			arr = null;
			}
        if (isFITS)
            {
            if (XnotY)
                {
                for (int ap = 0; ap < arr.length; ap++)
                    {
                    arr[ap] -= Centroid.PIXELCENTER;
                    }
                }
            else
                {
                for (int ap = 0; ap < arr.length; ap++)
                    {
                    arr[ap] = (double)imp.getHeight() - arr[ap] + Centroid.PIXELCENTER;
                    }                
                }
            }
        
		return arr;
		}
    
	/**
	 * Extracts a boolean array from a string array.
	 */
	protected boolean[] extractBoolean (String[] s)
		{
		boolean[] arr = new boolean[s.length];

        for (int i=0; i < arr.length; i++)
            {
            if (s[i].equalsIgnoreCase("true"))
                arr[i] = true;
            else
                arr[i] = false;
            }

		return arr;
		}    

	/**
	 * Stops reception of mouse and keyboard clicks.
	 */
	protected void noMoreInput()
		{
        if (!autoMode && helpFrame != null)
            {
            leftClickLabel.setText("");
            shiftLeftClickLabel.setText("");
            shiftControlLeftClickLabel.setText("");
            rightClickLabel.setText("");
            enterLabel.setText(""); 
            }
		canvas.removeMouseListener(this);
        canvas.removeKeyListener(this);
        if (imp.getWindow() instanceof astroj.AstroStackWindow)
            {
            aw = (AstroStackWindow)imp.getWindow();
            aw.setDisableShiftClick(false);
            }
		}

	/**
	 * Finishes whole process.
	 */
	public void shutDown()
		{
        noMoreInput();
		super.shutDown();
        cancelled=true;
        processingStack=false;
        stackTask = null;
        stackTaskTimer = null;
        doubleClickTask = null;
        doubleClickTaskTimer = null;
        if (table != null) table.setLock(false);
        closeHelpPanel();
        Prefs.set (MultiAperture_.PREFS_NAPERTURESMAX, nAperturesMax);
        Prefs.set (MultiAperture_.PREFS_PREVIOUS, previous);
        Prefs.set (MultiAperture_.PREFS_SINGLESTEP, singleStep);
        Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
        Prefs.set (MultiAperture_.PREFS_SHOWHELP, showHelp);
        Prefs.set (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
//        Prefs.set (MultiAperture_.PREFS_WIDETABLE, wideTable);
        Prefs.set (MultiAperture_.PREFS_SHOWRATIO, showRatio);
        Prefs.set (MultiAperture_.PREFS_SHOWCOMPTOT, showCompTot);
        Prefs.set (MultiAperture_.PREFS_SHOWRATIO_ERROR, showRatioError);
        Prefs.set (MultiAperture_.PREFS_SHOWRATIO_SNR, showRatioSNR);
        Prefs.set (MultiAperture_.PREFS_ENABLEDOUBLECLICKS, enableDoubleClicks);
//        Prefs.set (MultiAperture_.PREFS_FOLLOW, follow);
        Prefs.set (MultiAperture_.PREFS_AUTOMODE, "false");
        Prefs.set (MultiAperture_.PREFS_FINISHED, "true");
        Prefs.set (MultiAperture_.PREFS_USEMACROIMAGE, "false");
        Prefs.set (MultiAperture_.PREFS_XLOCATION, xLocation);
        Prefs.set (MultiAperture_.PREFS_YLOCATION, yLocation);
        Prefs.set ("plot2.helpFrameLocationX", helpFrameLocationX);
        Prefs.set ("plot2.helpFrameLocationY", helpFrameLocationY);  
		}

	//
	// MouseListener METHODS
	//

	/**
	 * Main MouseListener method used: process all mouse clicks.
	 */
	public void mouseReleased(MouseEvent e)
		{
        ee = e;
        if (autoMode || !enableDoubleClicks)
            {
            processSingleClick(ee);
            }
        else
            {
            if (e.getClickCount() == 1)
                {
                doubleClick = false;
                try	{
                    doubleClickTask = new TimerTask ()
                        {
                        public void run ()
                            {
                            if (!doubleClick) processSingleClick(ee);
                            doubleClickTask = null;
                            doubleClickTaskTimer = null;
                            }
                        };
                    doubleClickTaskTimer = new java.util.Timer();
                    if ((modis & InputEvent.BUTTON1_MASK) != 0)
                        doubleClickTaskTimer.schedule (doubleClickTask, 300);
                    else
                        doubleClickTaskTimer.schedule (doubleClickTask, 600);
                    }

                catch (Exception eee)
                    {
                    IJ.showMessage ("Error starting double click timer task : "+eee.getMessage());
                    }
                }
            else
                {
                doubleClick = true;
                return;
                }
            }
        }


    void processSingleClick(MouseEvent e)
        {
        if (!enterPressed && !autoMode)
            {
            screenX = e.getX();
            screenY = e.getY();
            modis = e.getModifiers();
            }

        //Right mouse click or <Enter> finalizes aperture selection or terminates single step mode
        if (enterPressed || ((Math.abs(screenX-startDragScreenX) + Math.abs(screenY-startDragScreenY) < 4.0 ) &&
                (modis & InputEvent.BUTTON3_MASK) != 0 && !e.isShiftDown() && !e.isControlDown() && !e.isAltDown()))
            {
            enterPressed = false;
            if (!aperturesInitialized)
                {
                if ((!previous && ngot > 0) || allStoredAperturesPlaced)
                    {
                    nApertures = ngot;
                    simulatedLeftClick = true;
                    aperturesInitialized = true;
                    }
                }
            else if (singleStep)
                {
                if (firstClick || allStoredAperturesPlaced)
                    {
                    IJ.beep();
                    shutDown();
                    }
                return;
                }
            }

        //do nothing unless automode or left mouse is clicked with no modifier keys pressed except "shift"
        if (autoMode || simulatedLeftClick || ((Math.abs(screenX-startDragScreenX) + Math.abs(screenY-startDragScreenY) < 4.0 ) &&
                (modis & InputEvent.BUTTON1_MASK) != 0 && 
                (!e.isControlDown() || e.isShiftDown()) && !e.isAltDown() && !e.isMetaDown()))
            {
            if (!autoMode && previous && !firstClick && !allStoredAperturesPlaced)
                {
                return;
                }
            
            simulatedLeftClick = false;
            
            if ( !autoMode && firstSlice > stackSize)
                {
                IJ.beep();
                shutDown();
                return;
                }
            if (frameAdvance)
                {
                firstSlice += 1;
                lastSlice = firstSlice;
                imp.setSlice(firstSlice);
                imp.updateImage();
                if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
                    {
//                    ocanvas = OverlayCanvas.getOverlayCanvas (img);
//                    canvas = ocanvas;
                    ocanvas.clearRois();
                    }
                ip = imp.getProcessor();
                firstClick = true;
                allStoredAperturesPlaced = false;
//                leftClickLabel.setText("Identify star 1 and place all apertures");
//                rightClickLabel.setText("Cancel Multi-Aperture");
//                enterLabel.setText("Cancel Multi-Aperture");                
                }
            else
                {
                slice = imp.getCurrentSlice();	

                // INITIAL CENTER FROM MOUSE CLICK OR PREVIOUS APERTURE LOCATION
                if (autoMode)
                    {
                    xCenter = xPos[0];
                    yCenter = yPos[0];
                    }
                else
                    {
                    if (e != null)
                        {
                        xCenter = canvas.offScreenXD(e.getX());
                        yCenter = canvas.offScreenYD(e.getY());
                        }
                    else
                        {
                        xCenter = 0;
                        yCenter = 0;
                        }
                    if (ngot < nApertures && !(previous && firstClick))
                        {
                        if ((!e.isShiftDown() && ngot>0) || (e.isShiftDown() && ngot==0))
                            {
                            setApertureColor(Color.red);
                            setApertureName("C"+(ngot+1));
                            isRefStar[ngot] = true;
                            }
                        else
                            {
                            setApertureColor(Color.green);
                            setApertureName("T"+(ngot+1));  
                            isRefStar[ngot] = false;
                            }                             
                            
                        apertureRemoved = ocanvas.removeRoi((int)xCenter, (int)yCenter);
                        if (!apertureRemoved && !measureAperture())
                            {
                            centerROI();
                            imp.unlock();
                            IJ.beep();
                            return;
                            }
                        }
                    }

                // ADD APERTURE TO LIST OR SHIFT OLD APERTURE POSITIONS

                if (autoMode)
                    {
                    if (isRefStar[ngot]) 
                        {
                        setApertureColor(Color.red);
                        setApertureName("C"+(ngot+1));
                        }
                    else
                        {
                        setApertureColor(Color.green);
                        setApertureName("T"+(ngot+1));  
                        }                    
                    if (!measureAperture())
                            {
                            centerROI();
                            imp.unlock();
                            IJ.beep();
                            return;
                            }
                    double dx = xCenter-xPos[0];
                    double dy = yCenter-yPos[0];
                    for (int i=0; i < nApertures; i++)
                        {
                        xPos[i] += dx;
                        yPos[i] += dy;
                        }
                    ngot = nApertures;
                    aperture = ngot;
                    }
                else if (previous && firstClick)
                    {
                    firstClick = false;
                    if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
                        {
                        ocanvas.clearRois();
                        }
                    double dx = xCenter-xPosStored[0];
                    double dy = yCenter-yPosStored[0];
                    for (int i=0; i < nAperturesStored; i++)
                        {
                        xPos[i] = xPosStored[i] + dx;
                        yPos[i] = yPosStored[i] + dy;
                        }
                    for (int i=0; i < nAperturesStored; i++)
                        {
                        xCenter = xPos[i];
                        yCenter = yPos[i];
                        if (isRefStar[i])
                            {
                            setApertureColor(Color.red);
                            setApertureName("C"+(i+1));
                            }
                        else
                            {
                            setApertureColor(Color.green);
                            setApertureName("T"+(i+1));
                            }
                        if (!measureAperture())
                            {
                            centerROI();
                            ngot = 0;
                            setApertureColor(Color.green);
                            setApertureName("T1");
                            imp.unlock();
                            IJ.beep();
                            return;
                            }      
                        if (i == 0)
                            {
                            double dxx = xCenter-xPos[0];
                            double dyy = yCenter-yPos[0];
                            for (int j=0; j < nApertures; j++)
                                { //update all positions based on new centroid of first clicked star
                                xPos[j] += dxx;
                                yPos[j] += dyy;
                                }
                            }
                        else
                            {  //update all comparison star positions after fine tuning the centroid of each individual comparison star
                            xPos[i] = xCenter;
                            yPos[i] = yCenter;
                            }
                        }
                    ngot = nAperturesStored;
                    aperture = ngot;
                    allStoredAperturesPlaced = true;
//                    leftClickLabel.setText("Add reference star aperture C"+(ngot+1)+", or delete aperture");
//                    shiftLeftClickLabel.setText("Add target star aperture T"+(ngot+1)+", or change T/C designation");
//                    shiftControlLeftClickLabel.setText("Add target star aperture T1, or rename aperture to T1");
//                    rightClickLabel.setText("Finalize aperture selection"+(singleStep?"":" and start processing"));
//                    enterLabel.setText("Finalize aperture selection"+(singleStep?"":" and start processing"));                     
                    }
                else if (!apertureRemoved && ngot < nApertures)
                    {
                    if (!e.isControlDown())
                        addAperture();
                    else
                        addApertureAsT1();
                    }
                else if (apertureRemoved)
                    {
                    if (!e.isShiftDown() &&!e.isControlDown())
                        removeAperture();
                    else if (e.isShiftDown() &&!e.isControlDown())
                        toggleApertureType();
                    else
                        renameApertureToT1();
                    apertureRemoved = false;
                    }
                }

            if (frameAdvance)
                {
                if (firstSlice > stackSize)
                    {
                    IJ.showStatus ("Finished processing stack. Click to exit.");
                    if (helpFrame != null) leftClickLabel.setText("End of stack, stop processing");
                    }
                else
                    {
                    IJ.showStatus ("Click to select first aperture (right click to exit).");
                    if (helpFrame != null)
                        {
                        leftClickLabel.setText("Identify star 1 and place all apertures");
                        shiftLeftClickLabel.setText("");
                        shiftControlLeftClickLabel.setText("");
                        rightClickLabel.setText("Cancel Multi-Aperture");
                        enterLabel.setText("Cancel Multi-Aperture");    
                        }
//                    rightClickLabel.setText("Finalize aperture selection"+(singleStep?"":" and start processing"));
//                    enterLabel.setText("Finalize aperture selection"+(singleStep?"":" and start processing"));                      
                    }
                frameAdvance = false;
                }
            else if (singleStep && ngot >= nApertures)
                {
                //PROCESS ONE SLICE AT A TIME WHILE IN SINGLE STEP MODE
                xOld = (double[]) xPos.clone();
                yOld = (double[]) yPos.clone();
                if (!checkResultsTable())
                    {
                    IJ.showMessage("Multi-Aperture failed to create Measurements table");
                    IJ.beep();
                    shutDown();
                    }
                processStack();
                frameAdvance = true;
                saveNewApertures ();
                previous = true;
                Prefs.set (MultiAperture_.PREFS_PREVIOUS, previous);
                IJ.showStatus ("Click to advance slice (right click to exit).");
                if (helpFrame != null)
                    {
                    leftClickLabel.setText("Advance to next image in stack");
                    shiftLeftClickLabel.setText("");
                    shiftControlLeftClickLabel.setText("");
                    rightClickLabel.setText("Cancel Multi-Aperture");
                    enterLabel.setText("Cancel Multi-Aperture"); 
                    }
                }

            // GOT ALL APERTURES?
            else if (ngot < nApertures)
                {
                infoMessage = "Click to select aperture #"+(ngot+1)+" (<ESC> to abort).";
                IJ.showStatus (infoMessage);
                if (helpFrame != null)
                    {
                    if (ngot > 0)
                        {
                        leftClickLabel.setText("Add reference star aperture C"+(ngot+1)+", or delete aperture");
                        shiftLeftClickLabel.setText("Add target star aperture T"+(ngot+1)+", or change T/C designation");
                        shiftControlLeftClickLabel.setText("Add target star aperture T1, or rename aperture to T1");
                        rightClickLabel.setText("Finalize aperture selection"+(singleStep?"":" and start processing"));
                        enterLabel.setText("Finalize aperture selection"+(singleStep?"":" and start processing"));   
                        }
                    else
                        {
                        leftClickLabel.setText("Add target star aperture T"+(ngot+1));
                        shiftLeftClickLabel.setText("Add reference star aperture C"+(ngot+1));
                        shiftControlLeftClickLabel.setText("");
                        rightClickLabel.setText("");
                        enterLabel.setText("");                    
                        }
                    }
                }
            else
                {	// PROCESS ALL SLICES WHEN NOT IN SINGLE STEP MODE
                noMoreInput ();
                xOld = (double[]) xPos.clone();
                yOld = (double[]) yPos.clone();
                saveNewApertures ();
                if (!checkResultsTable())
                    {
                    IJ.showMessage("Multi-Aperture failed to create Measurements table");
                    IJ.beep();
                    shutDown();
                    }
                if (!singleStep)
                    {
                    if (stackSize > 1 && doStack)
                        {
                        IJ.showStatus ("Processing stack...");
                        processingStack = true;
                        startProcessStack();
                        }
                    else
                        {
                        IJ.showStatus ("Processing image...");
                        processImage();
                        }
                    }
                if (!processingStack && !autoMode && (firstSlice < lastSlice)) IJ.beep();
                if (!processingStack) shutDown();
                }
            }
        }

        /**
	 * Saves new aperture locations to preferences.
	 */
        protected void saveNewApertures ()
            {
            String xpos = "";
            String ypos = "";
            String isref = "";
            for (int i=0; i < nApertures; i++)
                {
                if (i == 0)
                    {
                    xpos += (float)xPos[i];
                    ypos += (float)yPos[i];
                    isref += isRefStar[i];
                    }
                else
                    {
                    xpos += ","+(float)xPos[i];
                    ypos += ","+(float)yPos[i];
                    isref += ","+isRefStar[i];
                    }
                }
            if (aperturesInitialized)
                {
                IJ.showStatus("saving new aperture locations");
                xPosStored = xPos.clone();
                yPosStored = yPos.clone();
                nAperturesStored = nApertures;
                Prefs.set (MultiAperture_.PREFS_XAPERTURES, xpos);
                Prefs.set (MultiAperture_.PREFS_YAPERTURES, ypos);
                Prefs.set (MultiAperture_.PREFS_ISREFSTAR, isref);
                }
            }


        /**
	 * Adds the aperture parameters to the list of apertures.
	 */
	protected void addAperture ()
		{
		xPos[ngot] = xCenter;
        yPos[ngot] = yCenter;
        ngot++;
		aperture = ngot;
		}

    protected void addApertureAsT1()
        {
        for (int ap = ngot-1; ap >= 0; ap--)
            {
            xPos[ap+1] = xPos[ap];
            yPos[ap+1] = yPos[ap];
            isRefStar[ap+1] = isRefStar[ap];
            }
        xPos[0] = xCenter;
        yPos[0] = yCenter;
        isRefStar[0] = false;
        ngot++;

        ocanvas.clearRois();
		for (int ap=0;  ap < ngot; ap++)
			{
			aperture = ap;
            if (!isRefStar[ap])
                {
                setApertureColor(Color.green);
                setApertureName("T"+(ap+1));
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while adding a new aperture");
                shutDown();
                return;
				}
            }
        return;
        }

	protected void removeAperture ()
		{
        for (int ap = 0; ap < ngot; ap++)
            {
            if ((xCenter-xPos[ap])*(xCenter-xPos[ap])+(yCenter-yPos[ap])*(yCenter-yPos[ap]) <= radius*radius)
                {
                ngot--;
                for (int i = ap; i < ngot; i++)
                    {
                    xPos[i] = xPos[i+1];
                    yPos[i] = yPos[i+1];
                    isRefStar[i] = isRefStar[i+1];
                    }
                break;
                }
            }
        ocanvas.clearRois();
		for (int ap=0; ap < ngot; ap++)
			{
			aperture = ap;
            if (!isRefStar[ap])
                {
                setApertureColor(Color.green);
                setApertureName("T"+(ap+1));
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while deleting a different aperture");
                shutDown();
                return;
				}
            }
        return;
		}

    protected void renameApertureToT1()
        {
        for (int ap = 0; ap < ngot; ap++)
            {
            if ((xCenter-xPos[ap])*(xCenter-xPos[ap])+(yCenter-yPos[ap])*(yCenter-yPos[ap]) <= radius*radius)
                {
                for (int i = ap; i > 0; i--)
                    {
                    xPos[i] = xPos[i-1];
                    yPos[i] = yPos[i-1];
                    isRefStar[i] = isRefStar[i-1];
                    }
                break;
                }
            }
        xPos[0] = xCenter;
        yPos[0] = yCenter;
        isRefStar[0] = false;

        ocanvas.clearRois();
		for (int ap=0;  ap < ngot; ap++)
			{
			aperture = ap;
            if (!isRefStar[ap])
                {
                setApertureColor(Color.green);
                setApertureName("T"+(ap+1));
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while renaming an aperture to T1");
                shutDown();
                return;
				}
            }
        return;
        }

    protected void toggleApertureType()
        {
        for (int ap = 0; ap < ngot; ap++)
            {
            if ((xCenter-xPos[ap])*(xCenter-xPos[ap])+(yCenter-yPos[ap])*(yCenter-yPos[ap]) <= radius*radius)
                {
                isRefStar[ap] = !isRefStar[ap];
                break;
                }
            }

        ocanvas.clearRois();
		for (int ap=0;  ap < ngot; ap++)
			{
			aperture = ap;
            if (!isRefStar[ap])
                {
                setApertureColor(Color.green);
                setApertureName("T"+(ap+1));
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

            xCenter = xPos[ap];
            yCenter = yPos[ap];
			if (!measureAperture())
				{
                centerROI();
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while toggling aperture type");
                shutDown();
                return;
				}
            }
        return;
        }    

	public void mousePressed(MouseEvent e)
        {
        if (!autoMode)
            {
            startDragScreenX = e.getX();
            startDragScreenY = e.getY();
            }
        }
	public void mouseClicked(MouseEvent e) {}
	public void mouseExited(MouseEvent e)
			{
			IJ.showStatus (infoMessage);
			}
	public void mouseEntered(MouseEvent e) {}

        /** Handle the key typed event from the image canvas. */
    public void keyTyped(KeyEvent e) {

    }

    /** Handle the key-pressed event from the image canvas. */
    public void keyPressed(KeyEvent e) {
        int keyCode = e.getKeyCode();
        if (IJ.escapePressed())
            {
            IJ.beep();
            Prefs.set (MultiAperture_.PREFS_CANCELED, "true");
            shutDown();
            }
        else if (keyCode == KeyEvent.VK_ENTER)
            {
            enterPressed = true;
            processSingleClick(dummyClick);
            }
    }

    /** Handle the key-released event from the image canvas. */
    public void keyReleased(KeyEvent e) {

    }


	//
	// MultiAperture_ METHODS
	//

	/**
	 * Define the apertures and decide on the sub-stack if appropriate.
	 */
	protected boolean setUpApertures ()
		{
		// CHECK SLICES
        firstSlice = imp.getCurrentSlice();
		lastSlice = stackSize;
        if (singleStep)
                lastSlice = firstSlice;
        if (!autoMode)
                {
                GenericDialog gd = dialog();

                gd.showDialog();
                xLocation = gd.getX();
                yLocation = gd.getY();
                Prefs.set (MultiAperture_.PREFS_XLOCATION, xLocation);
                Prefs.set (MultiAperture_.PREFS_YLOCATION, yLocation);
                if (!gd.wasOKed())
                    {
                    cancelled = true;
                    return false;
                    }

			// GET UPDATED STANDARD PARAMETERS FROM REQUIRED DIALOG FIELDS:
			//	nApertures,firstSlice,lastSlice,previous,singleStep,oneTable,wideTable

			// NOTE: ONLY THE GENERIC MultiAperture_ FIELDS BELONG HERE !!!!!!!!!!!!!

                nAperturesMax = (int)gd.getNextNumber();
                nApertures = nAperturesMax;
                if (gd.invalidNumber() || nApertures <= 0)
                        {
                        IJ.beep();
                        IJ.error ("Invalid number of apertures: "+nApertures);
                        return false;
                        }
                Prefs.set (MultiAperture_.PREFS_NAPERTURESMAX, nAperturesMax);
                if (stackSize > 1)
                        {
                        firstSlice=(int)gd.getNextNumber();
                        if (gd.invalidNumber() || firstSlice < 1)
                                firstSlice=1;
                        lastSlice=(int)gd.getNextNumber();
                        if (gd.invalidNumber() || lastSlice > stackSize)
                                lastSlice= stackSize;
                        if (firstSlice != lastSlice)
                                {
                                if (firstSlice > lastSlice)
                                        {
                                        int i=firstSlice;
                                        firstSlice=lastSlice;
                                        lastSlice=i;
                                        }
                                doStack=true;
                                }
                        }
                else	{
                    firstSlice=1;
                    lastSlice=1;
                    }
                radius = gd.getNextNumber();
                if (gd.invalidNumber() || radius <= 1)
                        {
                        IJ.beep();
                        IJ.error ("Invalid aperture radius: "+radius);
                        return false;
                        }
                if (oldradius != radius)
                    changeAperture();
                rBack1 = gd.getNextNumber();
                if (gd.invalidNumber() || rBack1 < radius)
                        {
                        IJ.beep();
                        IJ.error ("Invalid background inner radius: "+rBack1);
                        return false;
                        }
                if (oldrBack1 != rBack1)
                    changeAperture();
                rBack2 = gd.getNextNumber();
                if (gd.invalidNumber() || rBack2 < rBack1)
                        {
                        IJ.beep();
                        IJ.error ("Invalid background outer radius: "+rBack2);
                        return false;
                        }
                if (oldrBack2 != rBack2)
                    changeAperture();
                previous = gd.getNextBoolean();
                singleStep = gd.getNextBoolean();
                if (singleStep)
                    lastSlice = firstSlice;
//                oneTable = !gd.getNextBoolean();
//                wideTable = gd.getNextBoolean();

			    Prefs.set (MultiAperture_.PREFS_PREVIOUS, previous);
			    Prefs.set (MultiAperture_.PREFS_SINGLESTEP, singleStep);
//                Prefs.set (MultiAperture_.PREFS_WIDETABLE, wideTable);

                // GET NON-STANDARD PARAMETERS AND CLEAN UP
			    return finishFancyDialog (gd);
                }

        return true;
		}

    protected void changeAperture()
        {
        apertureChanged = true;
        Prefs.set("setaperture.aperturechanged",apertureChanged);
		Prefs.set (Aperture_.AP_PREFS_RADIUS, radius);
		Prefs.set (Aperture_.AP_PREFS_RBACK1, rBack1);
		Prefs.set (Aperture_.AP_PREFS_RBACK2, rBack2);
        Prefs.set (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
        Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
        }


	protected void startProcessStack ()     //start new task to run stack processing so that table will update for each image processed
		{                                   //the mouse handler apparently blocks window updates to improve performance
		try	{
			stackTask = new TimerTask ()
				{
				public void run ()
					{
                    processStack();
                    stackTask = null;
                    stackTaskTimer = null;
                    }
                };
            stackTaskTimer = new java.util.Timer();
            stackTaskTimer.schedule (stackTask,0);
            }

        catch (Exception e)
            {
            IJ.showMessage ("Error starting process stack task : "+e.getMessage());
            }
        }


	/**
	 * Perform photometry on each image of selected sub-stack.
	 */
	protected void processStack ()
		{
		verbose=false;
//		vx = 0.0;
//		vy = 0.0;
		canvas = imp.getCanvas();
		ocanvas = null;
		for (int i=firstSlice; i <= lastSlice; i++)
			{
            slice=i;
			imp.setSlice(i);
			imp.updateImage();
			if (starOverlay || skyOverlay || valueOverlay || nameOverlay)
				{
				ocanvas = OverlayCanvas.getOverlayCanvas (imp);
				canvas = ocanvas;
				ocanvas.clearRois();
				}
			ip = imp.getProcessor();
            processImage();
			if (cancelled || IJ.escapePressed())
				{
				IJ.beep();
                Prefs.set (MultiAperture_.PREFS_CANCELED, "true");
                shutDown();
				return;
				}
			}
        if (processingStack)
            {
            IJ.beep();
            shutDown();
            return;
            }
        }

    static void checkAndLockTable()
        {
        if (table == null) return;
        int cnt = 0;  //timeout after 1 second
        while (table != null && table.isLocked() && cnt < 10)
            {
            IJ.wait(100);
            cnt++;
            }
        if (table != null) table.setLock(true); 
        }
    
    
	/**
	 * Perform photometry on each aperture of current image.
	 */
	protected void processImage ()
		{
		double dx = 0.0;		// CHANGE
		double dy = 0.0;

		double srcMean = 0.0;		// MEAN SOURCE BRIGHTNESSES AND BACKGROUNDS
		double bck = 0.0;
        
        checkAndLockTable();

        ratio = new double[nApertures];
        ratioError = new double[nApertures];
        ratioSNR = new double[nApertures];
//        isRefStar = new boolean [nApertures];
        src = new double[nApertures];
        srcVar = new double[nApertures];
        tot = new double[nApertures];
        totVar = new double[nApertures];
        
        if (!isInstanceOfStackAlign)
            {
            for (int r = 0; r < nApertures; r++)  //check for target star <--> ref star changes in table (i.e. changes from multi-plot)
                {
                if (table.getColumnIndex("Source-Sky_C"+(r+1)) != MeasurementTable.COLUMN_NOT_FOUND &&
                    table.getColumnIndex("Source-Sky_T"+(r+1)) == MeasurementTable.COLUMN_NOT_FOUND)
                    {
                    isRefStar[r] = true;
                    }
                else if (table.getColumnIndex("Source-Sky_T"+(r+1)) != MeasurementTable.COLUMN_NOT_FOUND &&
                         table.getColumnIndex("Source-Sky_C"+(r+1)) == MeasurementTable.COLUMN_NOT_FOUND)
                    {
                    isRefStar[r] = false;
                    }                               
                else
                    {
                    IJ.beep();
                    IJ.showMessage("Error checking aperture types in table");
                    }
                }
            }
        
        int totCcntAP = -1;
        for (int ap=0; ap < nApertures; ap++)
            {   
            if (!isRefStar[ap])
                {
                totCcntAP = ap;
                break;
                } 
            }
        for (int ap=0; ap < nApertures; ap++)
            {         
            ratio[ap] = 0.0;
            ratioError[ap] = 0.0;
            ratioSNR[ap] = 0.0;
            src[ap] = 0.0;
            srcVar[ap] = 0.0;
            tot[ap] = 0.0;
            totVar[ap] = 0.0;
            }
        xFWHM = 0.0;
        yFWHM = 0.0;

        if (useVarSizeAp)
            {
            setVariableAperture(false);
            for (int ap=0;  ap < nApertures; ap++)
                {
                aperture = ap;
                // GET POSITION ESTIMATE
                if (!isRefStar[ap])
                    {
                    setApertureColor(Color.green);
                    setApertureName("T"+(ap+1));
                    }
                else
                    {
                    setApertureColor(Color.red);
                    setApertureName("C"+(ap+1));
                    }
//                if (follow)
//                    {
//                    xCenter=0.5*(xPos[ap]+xOld[ap]+vxOld)+vx;
//                    yCenter=0.5*(yPos[ap]+yOld[ap]+vyOld)+vy;
//                    }
//                else
//                    {
                    xCenter = xPos[ap];
                    yCenter = yPos[ap];
//                    }

                // MEASURE NEW POSITION
//                centerROI();
                if (!adjustAperture())
                    {
                    centerROI();
                    setVariableAperture(false);
                    IJ.beep();
                    IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while measuring FWHM");
                    shutDown();
                    return;
                    }

                xOld[ap] = xPos[ap];
                yOld[ap] = yPos[ap];

                xPos[ap] = xCenter;		// STORE POSITION IN CASE IT DRIFTS WITHIN A STACK
                yPos[ap] = yCenter;

                if (ap==0)
                    {
                    double xDel = xPos[0] - xOld[0];
                    double yDel = yPos[0] - yOld[0];
                    for (int app=1; app < nApertures; app++)
                        {
                        xOld[app] = xPos[app];
                        yOld[app] = yPos[app];

                        xPos[app] += xDel;
                        yPos[app] += yDel;
                        }
                    }

                xFWHM += xWidth;
                yFWHM += yWidth;
                }

            setVariableAperture(true, Math.max(xFWHM/nApertures, yFWHM/nApertures)*apFWHMFactor, apFWHMFactor);
            }
        OverlayCanvas.getOverlayCanvas(imp).clearRois();
		for (int ap=0;  ap < nApertures; ap++)
			{
			aperture = ap;
            if (!isRefStar[ap])
                {
                setApertureColor(Color.green);
                setApertureName("T"+(ap+1));
                }
            else
                {
                setApertureColor(Color.red);
                setApertureName("C"+(ap+1));
                }

			// GET POSITION ESTIMATE

//			if (follow)
//				{
//				xCenter=0.5*(xPos[ap]+xOld[ap]+vxOld)+vx;
//				yCenter=0.5*(yPos[ap]+yOld[ap]+vyOld)+vy;
//				}
//			else
//				{
				xCenter = xPos[ap];
				yCenter = yPos[ap];
//				}

			// MEASURE NEW POSITION

//			centerROI();

			if (!measureAperture())
				{
                centerROI();
                setVariableAperture(false);
                IJ.beep();
                IJ.showMessage("Multi-Aperture: failed to find aperture "+ap+" centroid while measuring photometric values");
                shutDown();
                return;
				}

            // STORE RESULTS

			if (ap == 0)
				{
                suffix = isRefStar[ap] ? "_C1" : "_T1";
			 	storeResults();
				}
            else
				storeAdditionalResults (ap);

            // FOLLOW MOTION FROM FRAME TO FRAME

			dx += xCenter-xPos[ap];
			dy += yCenter-yPos[ap];

			xOld[ap] = xPos[ap];
			yOld[ap] = yPos[ap];

			xPos[ap]=xCenter;		// STORE POSITION IN CASE IT DRIFTS WITHIN A STACK
			yPos[ap]=yCenter;

			srcMean += source;
			bck += back;

            ratio[ap] = source;
            src[ap] = source;
            srcVar[ap] = serror*serror;
            if (isRefStar[ap])
                {
                for (int i=0; i<nApertures; i++)
                    {
                    if (i != ap)
                        {
                        tot[i] += source;
                        totVar[i] += srcVar[ap];
                        }
                    }
                }
            if (ap == 0)
                {
                double xDel = xPos[0] - xOld[0];
                double yDel = yPos[0] - yOld[0];
                for (int app=1; app < nApertures; app++)
                    {
                    xOld[app] = xPos[app];
                    yOld[app] = yPos[app];

                    xPos[app] += xDel;
                    yPos[app] += yDel;
                    }
                }

			// FOR DAUGHTER CLASSES....

			noteOtherApertureProperty (ap);
			}


        setVariableAperture(false);

		// COMPUTE APERTURE RATIO AND ERRORS AND UPDATE TABLE
        if (!isInstanceOfStackAlign && nApertures > 1)
            {
            if (showRatio)
                {
                for (int ap=0;  ap < nApertures; ap++)
                    {
                    if (tot[ap] == 0) 
                        ratio[ap] = 0;
                    else
                        ratio[ap] /= tot[ap];
                    table.addValue ("rel_flux_"+(isRefStar[ap]?"C":"T")+(ap+1), ratio[ap], 6);
                    if (showRatioError)
                        {
                        if (src[ap] == 0 || tot[ap] == 0)
                            table.addValue ("rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1), 0.0, 6);
                        else
                            table.addValue ("rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1), ratio[ap]*Math.sqrt(srcVar[ap]/(src[ap]*src[ap])+ totVar[ap]/(tot[ap]*tot[ap])), 6);
                        }
                    if (showRatioSNR)
                        {
                        if (src[ap] == 0 || tot[ap] == 0)
                            table.addValue ("rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1), 0.0, 6);
                        else
                            table.addValue ("rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1), 1/Math.sqrt(srcVar[ap]/(src[ap]*src[ap])+ totVar[ap]/(tot[ap]*tot[ap])), 6);
                        }
                    }
                }
            if (showCompTot)
                {
                table.addValue (TOTAL, totCcntAP < 0 ? 0.0 : tot[totCcntAP], 6);
                }
            if (showCompTot && showErrors)
                {
                table.addValue (TOTAL_ERROR, totCcntAP < 0 ? 0.0 : Math.sqrt(totVar[totCcntAP]), 6);
                }
            }
            
		// CALCULATE MEAN SHIFT, BRIGHTNESS, AND BACKGROUND

		xCenter = dx/nApertures;
		yCenter = dy/nApertures;
		source = srcMean/nApertures;
		back = bck/nApertures;

		// CALCULATE AND NOTE THE PIXEL VELOCITIES IN PIXELS/SLICE

//		if (follow)
//			{
//			vxOld = vx;
//			vyOld = vy;
//			vx = dx/nApertures;
//			vy = dy/nApertures;
//			table.addValue (VX, vx, 6);
//			table.addValue (VY, vy, 6);
//			}


		// UPDATE TABLE
        if (!isInstanceOfStackAlign)
            {        
            table.show();
            tablePanel = MeasurementTable.getTextPanel(tableName);
            if (tablePanel!=null)
                {
                int lastLine = tablePanel.getLineCount()-1;
                tablePanel.setSelection(lastLine, lastLine);
                tablePanel.resetSelection();
                }
            
            if (table != null) table.setLock(false);
            
            if ((updatePlot && !Data_Processor.active) || (Data_Processor.active && Data_Processor.runMultiPlot))  
                {
                if (MultiPlot_.mainFrame!=null)
                    {
                    if (MultiPlot_.getTable()!=null && MultiPlot_.getTable().equals(table))
                        {
//                        IJ.log("update plot");
                        MultiPlot_.updatePlot();
//                        IJ.log("update plot complete");
                        }
                    else
                        {
//                        IJ.log("setTable");
                        MultiPlot_.setTable(table, false);
//                        IJ.log("setTable complete");
                        }
                    }
                else
                    {
                    IJ.runPlugIn("MultiPlot_",tableName);
                    if (MultiPlot_.mainFrame!=null && MultiPlot_.getTable()!=null)
                        {
//                        IJ.log("setTable first time");
                        MultiPlot_.setTable(table, false);
//                        IJ.log("setTable first time complete");
                        }
                    }
                }
            }
        else
            {
            if (table != null) table.setLock(false);
            }
        }
    
    

	/**
	 * Stores results for additional apertures.
	 */
	void storeAdditionalResults (int ap)
		{
		if (isInstanceOfStackAlign || ap <= 0) return;

		String header = (isRefStar[ap]?"_C":"_T")+(ap+1);
		if (showPosition)
			{
			table.addValue (AP_XCENTER+header, xCenter, 6);
			table.addValue (AP_YCENTER+header, yCenter, 6);
			}
		if (showPositionFITS)
			{
			table.addValue (AP_XCENTER_FITS+header, xCenter + Centroid.PIXELCENTER, 6);
			table.addValue (AP_YCENTER_FITS+header, (double)imp.getHeight() - yCenter + Centroid.PIXELCENTER, 6);
			}        
		if (showRADEC && wcs != null && wcs.hasRaDec() && raDec != null)
			{
			table.addValue (AP_RA+header, raDec[0]/15.0, 6);
			table.addValue (AP_DEC+header, raDec[1], 6);
			}
		if (showPhotometry)
			{
			table.addValue (AP_SOURCE+header,   photom.sourceBrightness(), 6);
			if (showErrors) table.addValue (AP_SOURCE_ERROR+header,   photom.sourceError(), 6);
            if (showSNR) table.addValue (AP_SOURCE_SNR+header,   photom.sourceBrightness()/photom.sourceError(), 6);
			if (showBack) table.addValue (AP_BACK+header, photom.backgroundBrightness());
		  	if (showPeak) table.addValue (AP_PEAK+header, photom.peakBrightness(), 6);
            if (showMean) table.addValue (AP_MEAN+header, photom.meanBrightness(), 6);
			if (showSaturationWarning && photom.peakBrightness() > saturationWarningLevel &&
                photom.peakBrightness() > table.getValue(AP_WARNING, table.getCounter()-1))
                    {
                    table.setValue(AP_WARNING, table.getCounter()-1, photom.peakBrightness());
                    }
			}
		if (showWidths)
			{
			table.addValue (AP_XWIDTH+header,   xWidth, 6);
			table.addValue (AP_YWIDTH+header,   yWidth, 6);
			}
		if (showMeanWidth)
			table.addValue (AP_MEANWIDTH+header, 0.5*(xWidth+yWidth), 6);
        if (showAngle)
            table.addValue (AP_ANGLE+header, angle, 6);
        if (showRoundness)
            table.addValue (AP_ROUNDNESS+header, round, 6);
        if (showVariance)
            table.addValue (AP_VARIANCE+header, variance, 6);       
        
//		table.show();
		}


	/**
	 * Notes anything else which might be interesting about an aperture measurement.
	 */
	protected void noteOtherApertureProperty  (int ap)
		{
        if (isInstanceOfStackAlign) return;
		}
    
    static public void clearTable()
        {
        if (table != null && MeasurementTable.getTextPanel(tableName) != null)
            {
            table = new MeasurementTable(tableName);
            table.show();
            }
        else
            {
            IJ.showStatus("No table to clear");
            }
        if (table != null) table.setLock(false); 
        
        }

	/**
	 * Set up extended table format.
	 */
	protected boolean checkResultsTable ()
		{

        if (isInstanceOfStackAlign) return true;
        MeasurementTable plotTable = MultiPlot_.getTable();
        if (MultiPlot_.mainFrame!=null && plotTable != null && MeasurementTable.shorterName(plotTable.shortTitle()).equals("Measurements"))  
                {
                table = plotTable;
                }

        tablePanel = MeasurementTable.getTextPanel(tableName);

        if (table != null && tablePanel != null && table.getCounter() > 0) return true;  //!autoMode && ()

//		if (!super.checkResultsTable ()) return false;
        
	    table = MeasurementTable.getTable (tableName);

		if (table == null)
			{
			IJ.error ("Unable to open measurement table.");
			return false;
			}      

        checkAndLockTable();       

		int i=0;

        tablePanel = MeasurementTable.getTextPanel(tableName);
        
        if (showSliceNumber && table.getColumnIndex(AP_SLICE) == ResultsTable.COLUMN_NOT_FOUND)
            i=table.getFreeColumn (AP_SLICE);
        if (showSaturationWarning && table.getColumnIndex(AP_WARNING) == ResultsTable.COLUMN_NOT_FOUND)
            i=table.getFreeColumn (AP_WARNING);
		if (showTimes && table.getColumnIndex(AP_MJD) == ResultsTable.COLUMN_NOT_FOUND)
			i=table.getFreeColumn (AP_MJD);
		if (showFits && fitsKeywords != null)
			{
			String[] sarr = fitsKeywords.split(",");
			for (int l=0; l < sarr.length; l++)
				{
				if (!sarr[l].equals("") &&
						table.getColumnIndex(sarr[l]) == ResultsTable.COLUMN_NOT_FOUND)
					i=table.getFreeColumn (sarr[l]);
				}
			}
        if (showRadii)
            {
			if (table.getColumnIndex(AP_RSOURCE) == ResultsTable.COLUMN_NOT_FOUND)
				i=table.getFreeColumn (AP_RSOURCE);
            if (useVarSizeAp)
                {
                if (table.getColumnIndex(AP_FWHMMULT) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_FWHMMULT);                
                if (table.getColumnIndex(AP_BRSOURCE) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_BRSOURCE);
                }            
			if (table.getColumnIndex(AP_RBACK1) == ResultsTable.COLUMN_NOT_FOUND)
				i=table.getFreeColumn (AP_RBACK1);
			if (table.getColumnIndex(AP_RBACK2) == ResultsTable.COLUMN_NOT_FOUND)
				i=table.getFreeColumn (AP_RBACK2);
            }


		if (nApertures > 1)
			{
			if (showRatio)
				{
                for (int ap=0; ap < nApertures; ap++)
                    {
                    if (table.getColumnIndex("rel_flux_"+(isRefStar[ap]?"C":"T")+(ap+1)) == MeasurementTable.COLUMN_NOT_FOUND)
                        i=table.getFreeColumn ("rel_flux_"+(isRefStar[ap]?"C":"T")+(ap+1));
                    }
				if (showRatioError)
					{
                    for (int ap=0; ap < nApertures; ap++)
                        {
                        if (table.getColumnIndex("rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1)) == MeasurementTable.COLUMN_NOT_FOUND)
                            i=table.getFreeColumn ("rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1));
                        }
                    }
				if (showRatioSNR)
					{
                    for (int ap=0; ap < nApertures; ap++)
                        {
                        if (table.getColumnIndex("rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1)) == MeasurementTable.COLUMN_NOT_FOUND)
                            i=table.getFreeColumn ("rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1));
                        }
                    }
				}
            if (showCompTot)
                {
                if (table.getColumnIndex(TOTAL) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (TOTAL);
                if (showErrors && table.getColumnIndex(TOTAL_ERROR) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (TOTAL_ERROR);
                }
            }
            
        for (int ap=0; ap < nApertures; ap++)
            {
            String header = (isRefStar[ap]?"_C":"_T")+(ap+1);
            if (showPosition)
                {
                if (table.getColumnIndex( AP_XCENTER+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_XCENTER+header);
                if (table.getColumnIndex( AP_YCENTER+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_YCENTER+header);
                }
            if (showPositionFITS)
                {
                if (table.getColumnIndex( AP_XCENTER_FITS+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_XCENTER_FITS+header);
                if (table.getColumnIndex( AP_YCENTER_FITS+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_YCENTER_FITS+header);
                }                
            if (showRADEC && wcs != null && wcs.hasRaDec() && raDec != null)
                {
                if (table.getColumnIndex(AP_RA+header) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RA+header);
                if (table.getColumnIndex(AP_DEC+header) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_DEC+header);
                }
            if (showPhotometry)
                {
                if (table.getColumnIndex( AP_SOURCE+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_SOURCE+header);
                if (showErrors && table.getColumnIndex(AP_SOURCE_ERROR+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_SOURCE_ERROR+header);
                if (showSNR && table.getColumnIndex(AP_SOURCE_SNR+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_SOURCE_SNR+header);
                if (showPeak && table.getColumnIndex( AP_PEAK+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_PEAK+header);
                if (showMean && table.getColumnIndex( AP_MEAN+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_MEAN+header);                    
                if (showBack && table.getColumnIndex( AP_BACK+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_BACK+header);
                }
            if (showMeanWidth && table.getColumnIndex ( AP_MEANWIDTH) == MeasurementTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_MEANWIDTH+header);
            if (showWidths)
                {
                if (table.getColumnIndex( AP_XWIDTH+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_XWIDTH+header);
                if (table.getColumnIndex( AP_YWIDTH+header) == MeasurementTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn ( AP_YWIDTH+header);
                }
            if (showAngle && table.getColumnIndex(AP_ANGLE+header) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_ANGLE+header);
            if (showRoundness && table.getColumnIndex(AP_ROUNDNESS+header) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_ROUNDNESS+header);
            if (showVariance && table.getColumnIndex(AP_VARIANCE+header) == ResultsTable.COLUMN_NOT_FOUND)
                i=table.getFreeColumn (AP_VARIANCE+header);
            if (showRaw && isCalibrated)
                {
                if (table.getColumnIndex(AP_RAWSOURCE+header) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RAWSOURCE+header);
                if (table.getColumnIndex(AP_RAWBACK+header) == ResultsTable.COLUMN_NOT_FOUND)
                    i=table.getFreeColumn (AP_RAWBACK+header);
                }                
            }

        table.setLock(false);
		table.show();
		return true;
		}


	/**
	 * Add aperture number to overlay display.
	 */
	protected void addStringRoi (double x, double y, String text)
		{
		super.addStringRoi (x,y,"  #"+(aperture+1)+":  "+text.trim());
		}

	/**
	 * Standard preferences dialog for MultiAperture_
	 */
	protected GenericDialog dialog()
		{
        if (!Prefs.isLocationOnScreen(new Point(xLocation, yLocation)))
            {
            xLocation = 10;
            yLocation = 10;
            }

		GenericDialog gd = new GenericDialog("Multi-Aperture Measurements", xLocation, yLocation);
//		gd.addMessage ("Aperture radii should have been set with the \"Set Aperture\" tool (double-click icon).");

		gd.addNumericField ("   Maximum number of apertures per image :", nAperturesMax,0,6,"  (right click in image to finalize)");
        gd.addMessage("");
		if (stackSize > 1)
			{
            gd.addSlider("           First slice ", 1, stackSize, firstSlice == stackSize ? 1 : firstSlice);
            gd.addSlider("           Last slice ", 1, stackSize, lastSlice);
//			gd.addNumericField ("           First slice :", firstSlice == stackSize ? 1 : firstSlice,0);
//			gd.addNumericField ("           Last slice :",  lastSlice, 0);
	        }
        gd.addMessage("");
        gd.addSlider("Radius of object aperture", 1, radius>100?radius:100, radius);
        gd.addSlider("Inner radius of background annulus", 1, rBack1>100?rBack1:100, rBack1);
        gd.addSlider("Outer radius of background annulus", 1, rBack2>100?rBack2:100, rBack2);
//        gd.addNumericField ("Radius of object aperture",radius,2);
//		gd.addNumericField ("Inner radius of background annulus",rBack1,2);
//		gd.addNumericField ("Outer radius of background annulus",rBack2,2);
        gd.addMessage("");
		gd.addCheckbox ("Use previous "+nAperturesStored+" apertures (1-click to set first aperture location).",previous && nAperturesStored > 0);
		gd.addCheckbox ("Use single step mode (set new aperture locations for each image - right click to exit)",singleStep);
//		gd.addCheckbox ("Put results in image's own measurements table.", !oneTable);
//		gd.addCheckbox ("All measurements from one image on the same Measurements Table line.", wideTable);
		gd.addMessage (" ");

		// HERE ARE THE THINGS WHICH AREN'T ABSOLUTELY NECESSARY
		addFancyDialog (gd);

		// gd.addCheckbox ("Track moving object (experimental!).",follow);
		// gd.addMessage (" ");
		gd.addMessage ("PRESS \"OK\" AND SELECT APERTURE LOCATIONS WITH LEFT CLICKS. THEN RIGHT CLICK TO BEGIN PROCESSING.");
        gd.addMessage ("(to abort aperture selection or processing, press <ESC>)");
		return gd;
		}

	/**
	 * Adds options to MultiAperture_ dialog() which aren't absolutely necessary.
	 * Sub-classes of MultiAperture_ may choose to replace or extend this functionality if they use the original dialog().
	 */
	protected void addFancyDialog (GenericDialog gd)
		{
		// GET NON-REQUIRED DIALOGUE FIELDS:
		//	showRatio,showRatioError,showRatioSNR,useVarSizeAp,apFWHMFactor
        gd.addCheckbox ("Reposition aperture to object centroid (leave unchecked to center aperture at mouse click).", reposition);
        gd.addCheckbox ("Remove stars from background (i.e. pixels > 2 sigma from mean).", removeBackStars);
        gd.addCheckbox ("Assume background is a plane.", backIsPlane);
        gd.addCheckbox ("Update plot of measurements while running.", updatePlot);
		gd.addCheckbox ("Compute relative flux in all apertures (TN/Sum(CX) and CN/Sum(CX-CN)",showRatio);
		gd.addCheckbox ("Compute relative flux error (only if you check \"Compute relative flux\" above).",showRatioError);
		gd.addCheckbox ("Compute relative flux signal-to-noise (only if you check \"Compute relative flux\" above).",showRatioSNR);
        gd.addCheckbox ("Compute total comparison star counts.",showCompTot);
		gd.addCheckbox ("Vary photometer aperture radii based on FWHM.",useVarSizeAp);
		gd.addSlider ("FWHM multiplication factor :", 0.1, 5.1, apFWHMFactor);
        gd.addMessage (" ");
        gd.addCheckbox ("Allow left/right double click fast zoom-in/out (adds slight delay to aperture placement).", enableDoubleClicks);
        gd.addCheckbox ("Show help panel during aperture selection.", showHelp);
		}

	/**
	 * Last part of non-required dialog created by addFancyDialog().
	 * Sub-classes not using the original dialog() will need a dummy version of this method!
	 */
	protected boolean finishFancyDialog (GenericDialog gd)
		{
		// GET NON-REQUIRED DIALOGUE FIELDS:
		//	showRatio,showRatioError,showRatioSNR,useVarSizeAp,apFWHMFactor
        reposition      = gd.getNextBoolean();
        removeBackStars = gd.getNextBoolean();
        backIsPlane     = gd.getNextBoolean();
        updatePlot      = gd.getNextBoolean();
		showRatio       = gd.getNextBoolean();
        showCompTot     = gd.getNextBoolean();
		showRatioError  = gd.getNextBoolean();
		showRatioSNR    = gd.getNextBoolean();
		useVarSizeAp    = gd.getNextBoolean();
		apFWHMFactor    = gd.getNextNumber();
        enableDoubleClicks = gd.getNextBoolean();
        showHelp        = gd.getNextBoolean();
		if (gd.invalidNumber() || apFWHMFactor <= 0)
			{
            IJ.beep();
			IJ.error ("Invalid aperture FWHM factor");
            return false;
			}
        if (oldUseVarSizeAp != useVarSizeAp || oldapFWHMFactor != apFWHMFactor || oldRemoveBackStars != removeBackStars || oldBackIsPlane != backIsPlane)
            changeAperture();
		// follow = gd.getNextBoolean();
        Prefs.set (Aperture_.AP_PREFS_REPOSITION, reposition);
        Prefs.set (Aperture_.AP_PREFS_REMOVEBACKSTARS, removeBackStars);
        Prefs.set (Aperture_.AP_PREFS_BACKPLANE, backIsPlane);
		Prefs.set (MultiAperture_.PREFS_SHOWRATIO, showRatio);
        Prefs.set (MultiAperture_.PREFS_SHOWCOMPTOT, showCompTot);
		Prefs.set (MultiAperture_.PREFS_SHOWRATIO_ERROR, showRatioError);
		Prefs.set (MultiAperture_.PREFS_SHOWRATIO_SNR, showRatioSNR);
		Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, useVarSizeAp);
        Prefs.set (MultiAperture_.PREFS_SHOWHELP, showHelp);
		Prefs.set (MultiAperture_.PREFS_APFWHMFACTOR, apFWHMFactor);
        Prefs.set (MultiAperture_.PREFS_ENABLEDOUBLECLICKS, enableDoubleClicks);
        Prefs.set (MultiAperture_.PREFS_UPDATEPLOT, updatePlot);
        
        showHelpPanel();
        
		return true;
		}
    
    protected void showHelpPanel()
        {
        MAIcon = createImageIcon("astroj/images/multiaperture.png", "Multi-Aperture Icon");
        helpFrame = new JFrame ("Multi-Aperture Help");
        helpFrame.setIconImage(MAIcon.getImage());
        helpPanel = new JPanel (new SpringLayout());
        helpScrollPane = new JScrollPane(helpPanel);
        helpFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        helpFrame.addWindowListener(new WindowAdapter(){
            @Override
            public void windowClosing(WindowEvent e){
                    closeHelpPanel();
                    }});        
        JLabel leftClickName = new JLabel("left-Click:");
        leftClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(leftClickName);
        if (previous)
            leftClickLabel = new JLabel("Add previous stored apertures by clicking on location corresponding to the first star");
        else
            leftClickLabel = new JLabel("Add target star aperture T1");
        helpPanel.add(leftClickLabel);
        
        JLabel shiftLeftClickName = new JLabel("shift-left-Click:");
        shiftLeftClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(shiftLeftClickName);
        if (previous)
            shiftLeftClickLabel = new JLabel("");
        else
            shiftLeftClickLabel = new JLabel("Add reference star aperture C1");
        helpPanel.add(shiftLeftClickLabel);
        
        JLabel shiftControlLeftClickName = new JLabel("shift-control-left-Click:");
        shiftControlLeftClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(shiftControlLeftClickName);     
        if (previous)
            shiftControlLeftClickLabel = new JLabel("");
        else
            shiftControlLeftClickLabel = new JLabel("");
        helpPanel.add(shiftControlLeftClickLabel);
                
        JLabel rightClickName = new JLabel("right-Click:");
        rightClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(rightClickName);        
        rightClickLabel = new JLabel("");
        helpPanel.add(rightClickLabel);
        
        JLabel enterName = new JLabel("<enter>:");
        enterName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(enterName);        
        enterLabel = new JLabel("");
        helpPanel.add(enterLabel);        
        
        JLabel controlLeftClickName = new JLabel("control-left-Click:");
        controlLeftClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(controlLeftClickName);         
        controlLeftClickLabel = new JLabel("Zoom In");
        helpPanel.add(controlLeftClickLabel);
        
        JLabel controlRightClickName = new JLabel("control-right-Click:");
        controlRightClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(controlRightClickName);         
        controlRightClickLabel = new JLabel("Zoom Out");
        helpPanel.add(controlRightClickLabel);
                
        JLabel mouseWheelName = new JLabel("roll mouse wheel:");
        mouseWheelName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(mouseWheelName);         
        mouseWheelLabel = new JLabel("Zoom In/Out");
        helpPanel.add(mouseWheelLabel);
        
        JLabel leftClickDragName = new JLabel("left-click-drag:");
        leftClickDragName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(leftClickDragName);         
        leftClickDragLabel = new JLabel("Pan image up/down/left/right");
        helpPanel.add(leftClickDragLabel); 
        
        JLabel middleClickName = new JLabel("middle-click:");
        middleClickName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(middleClickName);         
        middleClickLabel = new JLabel("Center clicked point in image display (if enabled in Preferences menu)");
        helpPanel.add(middleClickLabel);        

        JLabel escapeName = new JLabel("<escape>:");
        escapeName.setHorizontalAlignment(JLabel.RIGHT);
        helpPanel.add(escapeName);         
        escapeLabel = new JLabel("Cancel Multi-Aperture");
        helpPanel.add(escapeLabel);   
        
        SpringUtil.makeCompactGrid (helpPanel, helpPanel.getComponentCount()/2, 2, 6,6,6,6);

        helpFrame.add (helpScrollPane);
        helpFrame.pack();
        helpFrame.setResizable (true);
        Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
        if (!Prefs.isLocationOnScreen(new Point(helpFrameLocationX,helpFrameLocationY)))
            {
            helpFrameLocationX = mainScreenSize.width/2 - helpFrame.getWidth()/2;
            helpFrameLocationY = mainScreenSize.height/2 - helpFrame.getHeight()/2;
            }
        helpFrame.setLocation(helpFrameLocationX, helpFrameLocationY);
        helpFrame.setVisible (showHelp);        
        }
    
        /** Returns an ImageIcon, or null if the path was invalid. */
        protected ImageIcon createImageIcon(String path, String description) {
            MultiPlot_ m = new MultiPlot_();
            java.net.URL imgURL = m.getClass().getResource(path);
            if (imgURL != null) {
                return new ImageIcon(imgURL, description);
            } else {
                IJ.log("Couldn't find icon file: " + path);
                return null;
            }
        }    
    
        protected void closeHelpPanel()
            {
            if (helpFrame != null && helpFrame.isShowing())
                {
                helpFrameLocationX = helpFrame.getLocation().x;
                helpFrameLocationY = helpFrame.getLocation().y;
                helpFrame.setVisible(false);
                Prefs.set("plot2.helpFrameLocationX",helpFrameLocationX);
                Prefs.set("plot2.helpFrameLocationY",helpFrameLocationY);
                }
            if (helpFrame != null)
                {
                helpFrame.dispose();
                helpFrame = null;
                }
            }    
    
	}

